# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from avenue import Avenue
from avenue import MethodProcessor
from avenue import NotFound
from avenue import Skip
from pytest import raises


class TestAvenue(object):
    def test_basics(self):
        router = Avenue()

        @router.attach(path='/', method='GET')
        @router.attach(path='/welcome')
        def route_1():
            return 'Route 1'

        @router.attach(path='/', method='POST')
        def route_2():
            return 'Route 2'

        @router.attach(path='/post/<id|int>')
        def route_3(id):
            return 'Route 3 - for {:s}'.format(repr(id))

        @router.attach(path='/post/new')
        def route_4():
            return 'Route 4'

        @router.attach(path='/<page|greedy>')
        def route_5(page):
            return 'Route 5 - fallback - {:s}'.format(page)

        routing = [
            [{'path': '/', 'method': 'GET'},
             'Route 1'],
            [{'path': '/', 'method': 'POST'},
             'Route 2'],
            [{'path': '/post/12', 'method': 'GET'},
             'Route 3 - for 12'],
            [{'path': '/post/new-article-here', 'method': 'GET'},
             'Route 5 - fallback - post/new-article-here'],
            [{'path': '/post/new', 'method': 'GET'},
             'Route 4'],
            [{'path': '/any/other/page', 'method': 'GET'},
             'Route 5 - fallback - any/other/page'],
        ]

        for route, result in routing:
            assert router.solve(**route) == result

    def test_ordering(self):
        router = Avenue()

        @router.attach(path='/')
        def route_1():
            return 'Route 1'

        @router.attach(path='/', method='GET')
        def route_2():
            return 'Route 2'

        assert router.solve(path='/') == 'Route 1'
        assert router.solve(path='/', method='GET') == 'Route 2'

    def test_wrap(self):
        def wrap(func):
            def inner(**kwargs):
                return 'Wrapped: ' + func(**kwargs)

            return inner

        router = Avenue()

        @router.attach(path='/')
        def route():
            return 'Route 1'

        assert router.solve(path='/', wrap=wrap) == 'Wrapped: Route 1'

    def test_skipping(self):
        router = Avenue()
        toggle = 0

        @router.attach(path='/')
        def route_1():
            if toggle == 0:
                return 'Route 1'
            raise Skip()

        @router.attach(path='/')
        def route_2():
            return 'Route 2'

        assert router.solve(path='/') == 'Route 1'

        toggle = 1
        assert router.solve(path='/') == 'Route 2'

    def test_missing(self):
        router = Avenue()

        @router.attach(path='/', method='GET')
        @router.attach(path='/welcome')
        def route_1():
            return 'Route 1'

        with raises(NotFound):
            router.solve(path='/does/not/exist', method='GET')

        with raises(NotFound):
            router.solve()

        router = Avenue([MethodProcessor('method', optional=True)])

        @router.attach()
        def route_2():
            return 'Route 2'

        assert router.solve() == 'Route 2'



from __future__ import absolute_import, division, print_function

from .avenue import (
    Avenue,
    AvenueException,
    NotFound,
    PROCESSORS,
    Skip,
)

from .processors import (
    MatchError,
    MethodProcessor,
    PathProcessor,
    Processor,
)

__version__ = '0.2.0'

__package__ = 'avenue'
__title__ = 'Avenue'
__description__ = 'Avenue: Highway routing.'
__uri__ = 'https://github.com/CorverDevelopment/Avenue/'

__author__ = 'Nils Corver'
__email__ = 'nils@corverdevelopment.nl'

__license__ = 'MIT'
__copyright__ = 'Copyright (c) 2016 Corver Development B.V.'

__all__ = [
    'Avenue',
    'AvenueException',
    'NotFound',
    'PROCESSORS',
    'Skip',

    'MatchError',
    'MethodProcessor',
    'PathProcessor',
    'Processor',
]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from avenue import Avenue
from avenue import MatchError
from avenue import MethodProcessor
from avenue import PathProcessor
from avenue import Processor
from avenue.processors import Match
from pytest import raises


class TestProcessors(object):
    def test_match(self):
        match = Match()
        match.update(2)
        match.update(1)

        assert repr(match) == '<Match (score=[2, 1])>'

    def test_path_processor(self):
        avenue = Avenue([
            PathProcessor('path_1'),
            PathProcessor('path_2', optional=True),
        ])

        @avenue.attach(path_1='/welcome', path_2='/nl')
        def nl_welcome():
            return 'Welkom.'

        @avenue.attach(path_1='/welcome', path_2='/de-DE')
        def de_welcome():
            return 'Willkommen.'

        @avenue.attach(path_1='/welcome')
        def welcome():
            return 'Welcome.'

        assert avenue.solve(path_1='/welcome',
                            path_2='/de-DE') == 'Willkommen.'

        assert avenue.solve(path_1='/welcome') == 'Welcome.'

    def test_exceptions(self):
        avenue = Avenue()

        @avenue.attach()
        def will_never_run():
            pass

        with raises(RuntimeError):
            avenue.solve()

    def test_custom_processor(self):
        class RoleProcessor(Processor):
            def __call__(self, value, route):
                # prepare takes care of the `optional` part automaticly
                argument, value = self.prepare(value, route)

                # argument is the parameter with which the route was defined
                # value is the parameter which `solve` was called with

                print(value, 'in', argument)
                if not argument:
                    # processor is optional
                    return 0
                elif value in argument:
                    # value is one of the allowed roles
                    return -1
                else:
                    # you could also raise some other error and return
                    # a 403 to your users, this simply marks the route
                    # as not runnable if the role does not match
                    error = 'Role does not match, `{:s}` not in `{:s}`'.format(
                        value, ', '.join(argument))
                    raise MatchError(error)

        # create a router with the RoleProcessor
        router = Avenue([
            PathProcessor('path'),
            MethodProcessor('method', optional=True),
            RoleProcessor('role', optional=True),
        ])

        # define a route without a role
        @router.attach(path='/', method='GET')
        def hello_world():
            return 'Hallo world!'

        # define a route and use the role parameter
        @router.attach(path='/', method='GET', role=['user', 'administrator'])
        def hello_world_for_authenticated_people():
            return 'Hallo world, for users and administrators!'

        # define a route that matches the path, but accepts other roles
        @router.attach(path='/', method='GET', role=['spy'])
        def hello_world_for_spies():
            return 'Hallo world, nothing to see here!'

        # solve without a role
        route = {'path': '/', 'method': 'GET'}
        assert router.solve(**route) == 'Hallo world!'

        # solve with a role
        route = {'path': '/', 'method': 'GET', 'role': 'user'}
        assert router.solve(
            **route) == 'Hallo world, for users and administrators!'

        # solve with the spy role
        route = {'path': '/', 'method': 'GET', 'role': 'spy'}
        assert router.solve(
            **route) == 'Hallo world, nothing to see here!', router.solve(
            **route)



from __future__ import absolute_import, division, print_function

from pandora import CONVERTERS
from pandora import Path


class Match(object):
    def __init__(self, score=None, kwargs=None):
        self.score = score or list()
        self.kwargs = kwargs or dict()

    def update(self, result):
        if type(result) is int:
            self.score.append(result)
        elif isinstance(result, Match):
            self.score.append(result.score)
            self.kwargs.update(result.kwargs)

    def __lt__(self, other):
        return any([s < other.score[sx] for sx, s in enumerate(self.score)])

    def __repr__(self):
        return '<Match (score={:s})>'.format(repr(self.score))


class MatchError(Exception):
    pass


class Processor(object):
    def __init__(self, name, optional=False):
        self.name = name
        self.optional = optional

    def prepare(self, value, route):
        argument = route.kwargs.get(self.name)

        if not argument and not self.optional:
            message = 'Cannot run route, `{:s}` is not optional.'.format(
                self.name)

            raise RuntimeError(message)

        if argument and value is None:
            message = 'Value not found, `{:s}` is not optional.'.format(
                self.name)

            raise MatchError(message)

        return argument, value


class PathProcessor(Processor):
    def __init__(self, name, optional=False, converters=None):
        super(PathProcessor, self).__init__(name, optional)
        self.cache = dict()
        self.converters = converters or CONVERTERS

    def __call__(self, value, route):
        """

        :type value: mixed
        :type route: avenue.Route
        :rtype: Match
        """
        argument, value = self.prepare(value, route)

        if not argument:
            return True, -100, 100

        if argument not in self.cache:
            self.cache[argument] = Path(argument, self.converters,
                                        split='/')

        path = self.cache[argument]
        match = path(value)

        if match is not None:
            return Match(path, match)

        raise MatchError('Path does not match, `{:s}` != `{:s}`'.format(
            path.original, argument))


class MethodProcessor(Processor):
    def __call__(self, value, route):
        """

        :type value: mixed
        :type route: avenue.Route
        :rtype: Match
        """
        argument, value = self.prepare(value, route)

        if not argument:
            return 0
        elif value == argument:
            return -1
        else:
            error = 'Method does not match, `{:s}` != `{:s}`'.format(
                value, argument)
            raise MatchError(error)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from .processors import Match
from .processors import MatchError
from .processors import MethodProcessor
from .processors import PathProcessor
import logging
import operator

PROCESSORS = [
    PathProcessor('path'),
    MethodProcessor('method', optional=True),
]

logger = logging.getLogger('avenue')


class AvenueException(Exception):
    pass


class Skip(AvenueException):
    pass


class NotFound(AvenueException):
    pass


class Avenue(object):
    def __init__(self, processors=None):
        self.routes = list()
        self.processors = processors or PROCESSORS

    def attach(self, **kwargs):
        def inner(func):
            self.routes.append(Route(self, kwargs, func))
            return func

        return inner

    def match(self, **kwargs):
        values = [[p, kwargs.get(p.name)] for p in self.processors]
        matches = list()

        for route in self.routes:
            try:
                match = Match()
                for processor, value in values:
                    match.update(processor(value, route))
            except MatchError:
                pass
            else:
                matches.append([route, match])

        if len(matches):
            size = max(map(len, [m[1].score for m in matches]))
            matches = list(filter(lambda m: len(m[1].score) == size, matches))
            matches.sort(key=operator.itemgetter(1))

        logger.debug('Matched %s', [v[1] for v in values])
        for match in matches:
            logger.debug('%s', match)

        return matches

    def solve(self, **kwargs):
        wrap = kwargs.pop('wrap', None)

        for path, match in self.match(**kwargs):
            try:
                if callable(wrap):
                    callback = wrap(path.func)
                else:
                    callback = path.func

                return callback(**match.kwargs)
            except Skip:
                logger.debug('Callback raised Skip for match %s', match)

        raise NotFound()


class Route(object):
    def __init__(self, avenue, kwargs, func):
        self.avenue = avenue
        self.kwargs = kwargs
        self.func = func



from assemble import get_package

package = get_package()

keywords = [
    "routing"
]
classifiers = [
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Developers",
    "Natural Language :: English",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",

    "Programming Language :: Python",
    "Programming Language :: Python :: 2",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: Implementation :: CPython",

    "Topic :: Software Development :: Libraries :: Python Modules",
]

if __name__ == "__main__":
    package.setup(keywords, classifiers)
